@@ -1,5 +1,5 @@
 #############################################################################
-# Copyright (c) 2018, Voila Contributors                                    #
+# Copyright (c) 2018, Voilà Contributors                                    #
 # Copyright (c) 2018, QuantStack                                            #
 #                                                                           #
 # Distributed under the terms of the BSD 3-Clause License.                  #
@@ -7,10 +7,9 @@
 # The full license is in the file LICENSE, distributed with this software.  #
 #############################################################################
 
-from zmq.eventloop import ioloop
-
 import gettext
 import io
+import sys
 import json
 import logging
 import threading
@@ -39,12 +38,12 @@ from traitlets.config.application import Application
 from traitlets.config.loader import Config
 from traitlets import Unicode, Integer, Bool, Dict, List, default
 
-from jupyter_server.services.kernels.kernelmanager import MappingKernelManager
+from jupyter_server.services.kernels.kernelmanager import AsyncMappingKernelManager
 from jupyter_server.services.kernels.handlers import KernelHandler, ZMQChannelsHandler
 from jupyter_server.services.contents.largefilemanager import LargeFileManager
 from jupyter_server.base.handlers import FileFindHandler, path_regex
 from jupyter_server.config_manager import recursive_update
-from jupyter_server.utils import url_path_join
+from jupyter_server.utils import url_path_join, run_sync
 from jupyter_server.services.config import ConfigManager
 
 from jupyter_client.kernelspec import KernelSpecManager
@@ -53,17 +52,15 @@ from jupyter_core.paths import jupyter_config_path, jupyter_path
 
 from ipython_genutils.py3compat import getcwd
 
-from .paths import ROOT, STATIC_ROOT, collect_template_paths
+from .paths import ROOT, STATIC_ROOT, collect_template_paths, collect_static_paths
 from .handler import VoilaHandler
 from .treehandler import VoilaTreeHandler
 from ._version import __version__
-from .static_file_handler import MultiStaticFileHandler, WhiteListFileHandler
+from .static_file_handler import MultiStaticFileHandler, TemplateStaticFileHandler, WhiteListFileHandler
 from .configuration import VoilaConfiguration
-from .execute import VoilaExecutePreprocessor
+from .execute import VoilaExecutor
 from .exporter import VoilaExporter
-from .csspreprocessor import VoilaCSSPreprocessor
 
-ioloop.install()
 _kernel_id_regex = r"(?P<kernel_id>\w+-\w+-\w+-\w+-\w+)"
 
 
@@ -77,7 +74,10 @@ class Voila(Application):
     examples = 'voila example.ipynb --port 8888'
 
     flags = {
-        'debug': ({'Voila': {'log_level': logging.DEBUG}}, _("Set the log level to logging.DEBUG")),
+        'debug': (
+            {'Voila': {'log_level': logging.DEBUG, 'show_tracebacks': True}},
+            _("Set the log level to logging.DEBUG, and show exception tracebacks in output.")
+        ),
         'no-browser': ({'Voila': {'open_browser': False}}, _('Don\'t open the notebook in a browser after startup.'))
     }
 
@@ -98,7 +98,7 @@ class Voila(Application):
         8866,
         config=True,
         help=_(
-            'Port of the voila server. Default 8866.'
+            'Port of the Voilà server. Default 8866.'
         )
     )
     autoreload = Bool(
@@ -129,9 +129,8 @@ class Voila(Application):
     }
     classes = [
         VoilaConfiguration,
-        VoilaExecutePreprocessor,
-        VoilaExporter,
-        VoilaCSSPreprocessor
+        VoilaExecutor,
+        VoilaExporter
     ]
     connection_dir_root = Unicode(
         config=True,
@@ -146,7 +145,7 @@ class Voila(Application):
         '/',
         config=True,
         help=_(
-            'Path for voila API calls. If server_url is unset, this will be \
+            'Path for Voilà API calls. If server_url is unset, this will be \
             used for both the base route of the server and the client. \
             If server_url is set, the server will server the routes prefixed \
             by server_url, while the client will prefix by base_url (this is \
@@ -159,7 +158,7 @@ class Voila(Application):
         config=True,
         allow_none=True,
         help=_(
-            'Path to prefix to voila API handlers. Leave unset to default to base_url'
+            'Path to prefix to Voilà API handlers. Leave unset to default to base_url'
         )
     )
 
@@ -168,24 +167,15 @@ class Voila(Application):
         config=True,
         allow_none=True,
         help=_(
-            'path to notebook to serve with voila'
-        )
-    )
-
-    nbconvert_template_paths = List(
-        [],
-        config=True,
-        help=_(
-            'path to nbconvert templates'
+            'path to notebook to serve with Voilà'
         )
     )
 
     template_paths = List(
         [],
-        allow_none=True,
         config=True,
         help=_(
-            'path to nbconvert templates'
+            'path to jinja2 templates'
         )
     )
 
@@ -197,6 +187,10 @@ class Voila(Application):
         )
     )
 
+    show_tracebacks = Bool(False, config=True, help=_(
+        'Whether to send tracebacks to clients on exceptions.'
+    ))
+
     port_retries = Integer(50, config=True,
                            help=_("The number of additional ports to try if the specified port is not available.")
                            )
@@ -324,7 +318,35 @@ class Voila(Application):
         else:
             return getcwd()
 
+    def _init_asyncio_patch(self):
+        """set default asyncio policy to be compatible with tornado
+        Tornado 6 (at least) is not compatible with the default
+        asyncio implementation on Windows
+        Pick the older SelectorEventLoopPolicy on Windows
+        if the known-incompatible default policy is in use.
+        do this as early as possible to make it a low priority and overrideable
+        ref: https://github.com/tornadoweb/tornado/issues/2608
+        FIXME: if/when tornado supports the defaults in asyncio,
+               remove and bump tornado requirement for py38
+        """
+        if sys.platform.startswith("win") and sys.version_info >= (3, 8):
+            import asyncio
+            try:
+                from asyncio import (
+                    WindowsProactorEventLoopPolicy,
+                    WindowsSelectorEventLoopPolicy,
+                )
+            except ImportError:
+                pass
+                # not affected
+            else:
+                if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
+                    # WindowsProactorEventLoopPolicy is not compatible with tornado 6
+                    # fallback to the pre-3.8 default of Selector
+                    asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())
+
     def initialize(self, argv=None):
+        self._init_asyncio_patch()
         self.log.debug("Searching path %s for config files", self.config_file_paths)
         # to make config_file_paths settable via cmd line, we first need to parse it
         super(Voila, self).initialize(argv)
@@ -350,14 +372,10 @@ class Voila(Application):
 
     def setup_template_dirs(self):
         if self.voila_configuration.template:
-            collect_template_paths(
-                self.nbconvert_template_paths,
-                self.static_paths,
-                self.template_paths,
-                self.voila_configuration.template)
-            # look for possible template-related config files
-            template_conf_dir = [os.path.join(k, '..') for k in self.nbconvert_template_paths]
-            conf_paths = [os.path.join(d, 'conf.json') for d in template_conf_dir]
+            template_name = self.voila_configuration.template
+            self.template_paths = collect_template_paths(['voila', 'nbconvert'], template_name, prune=True)
+            self.static_paths = collect_static_paths(['voila', 'nbconvert'], template_name)
+            conf_paths = [os.path.join(d, 'conf.json') for d in self.template_paths]
             for p in conf_paths:
                 # see if config file exists
                 if os.path.exists(p):
@@ -367,10 +385,9 @@ class Voila(Application):
                     # update the overall config with it, preserving CLI config priority
                     if 'traitlet_configuration' in conf:
                         recursive_update(conf['traitlet_configuration'], self.voila_configuration.config.VoilaConfiguration)
-                        # pass merged config to overall voila config
+                        # pass merged config to overall Voilà config
                         self.voila_configuration.config.VoilaConfiguration = Config(conf['traitlet_configuration'])
         self.log.debug('using template: %s', self.voila_configuration.template)
-        self.log.debug('nbconvert template paths:\n\t%s', '\n\t'.join(self.nbconvert_template_paths))
         self.log.debug('template paths:\n\t%s', '\n\t'.join(self.template_paths))
         self.log.debug('static paths:\n\t%s', '\n\t'.join(self.static_paths))
         if self.notebook_path and not os.path.exists(self.notebook_path):
@@ -392,7 +409,7 @@ class Voila(Application):
             parent=self
         )
 
-        self.kernel_manager = MappingKernelManager(
+        self.kernel_manager = AsyncMappingKernelManager(
             parent=self,
             connection_dir=self.connection_dir,
             kernel_spec_manager=self.kernel_spec_manager,
@@ -442,14 +459,18 @@ class Voila(Application):
         handlers.extend([
             (url_path_join(self.server_url, r'/api/kernels/%s' % _kernel_id_regex), KernelHandler),
             (url_path_join(self.server_url, r'/api/kernels/%s/channels' % _kernel_id_regex), ZMQChannelsHandler),
+            (
+                url_path_join(self.server_url, r'/voila/templates/(.*)'),
+                TemplateStaticFileHandler
+            ),
             (
                 url_path_join(self.server_url, r'/voila/static/(.*)'),
                 MultiStaticFileHandler,
                 {
                     'paths': self.static_paths,
                     'default_filename': 'index.html'
-                }
-            )
+                },
+            ),
         ])
 
         # Serving notebook extensions
@@ -485,7 +506,7 @@ class Voila(Application):
                 VoilaHandler,
                 {
                     'notebook_path': os.path.relpath(self.notebook_path, self.root_dir),
-                    'nbconvert_template_paths': self.nbconvert_template_paths,
+                    'template_paths': self.template_paths,
                     'config': self.config,
                     'voila_configuration': self.voila_configuration
                 }
@@ -499,7 +520,7 @@ class Voila(Application):
                 (url_path_join(self.server_url, r'/voila/render/(.*)'),
                  VoilaHandler,
                  {
-                     'nbconvert_template_paths': self.nbconvert_template_paths,
+                     'template_paths': self.template_paths,
                      'config': self.config,
                      'voila_configuration': self.voila_configuration
                  }),
@@ -510,7 +531,7 @@ class Voila(Application):
 
     def stop(self):
         shutil.rmtree(self.connection_dir)
-        self.kernel_manager.shutdown_all()
+        run_sync(self.kernel_manager.shutdown_all())
 
     def random_ports(self, port, n):
         """Generate a list of n random ports near the given port.
@@ -528,7 +549,7 @@ class Voila(Application):
             try:
                 self.app.listen(port)
                 self.port = port
-                self.log.info('Voila is running at:\n%s' % self.display_url)
+                self.log.info('Voilà is running at:\n%s' % self.display_url)
             except socket.error as e:
                 if e.errno == errno.EADDRINUSE:
                     self.log.info(_('The port %i is already in use, trying another port.') % port)
@@ -544,7 +565,7 @@ class Voila(Application):
                 break
 
         if not success:
-            self.log.critical(_('ERROR: the voila server could not be started because '
+            self.log.critical(_('ERROR: the Voilà server could not be started because '
                               'no available port could be found.'))
             self.exit(1)
 
