@@ -1,6 +1,6 @@
 #############################################################################
 # Copyright (c) 2018, QuantStack                                            #
-# Copyright (c) 2018, Voila Contributors                                    #
+# Copyright (c) 2018, Voilà Contributors                                    #
 #                                                                           #
 # Distributed under the terms of the BSD 3-Clause License.                  #
 #                                                                           #
@@ -8,13 +8,19 @@
 #############################################################################
 
 import traitlets.config
-from traitlets import Unicode, Bool, Dict, List
+from traitlets import Unicode, Bool, Dict, List, Int, Enum
 
 
 class VoilaConfiguration(traitlets.config.Configurable):
     """Common configuration options between the server extension and the application."""
+    allow_template_override = Enum(['YES', 'NOTEBOOK', 'NO'], 'YES', help='''
+    Allow overriding the template (YES), or not (NO), or only from the notebook metadata.
+    ''').tag(config=True)
+    allow_theme_override = Enum(['YES', 'NOTEBOOK', 'NO'], 'YES', help='''
+    Allow overriding the theme (YES), or not (NO), or only from the notebook metadata.
+    ''').tag(config=True)
     template = Unicode(
-        'default',
+        'lab',
         config=True,
         allow_none=True,
         help=(
@@ -31,7 +37,7 @@ class VoilaConfiguration(traitlets.config.Configurable):
     ).tag(config=True)
     theme = Unicode('light').tag(config=True)
     strip_sources = Bool(True, help='Strip sources from rendered html').tag(config=True)
-    enable_nbextensions = Bool(False, config=True, help=('Set to True for Voila to load notebook extensions'))
+    enable_nbextensions = Bool(False, config=True, help=('Set to True for Voilà to load notebook extensions'))
 
     file_whitelist = List(
         Unicode(),
@@ -70,3 +76,8 @@ class VoilaConfiguration(traitlets.config.Configurable):
         --VoilaConfiguration.extension_language_mapping='{".py": "python", ".cpp": "C++11"}'
         ''',
     ).tag(config=True)
+
+    http_keep_alive_timeout = Int(10, help="""
+    When a cell takes a long time to execute, the http connection can timeout (possibly because of a proxy).
+    Voila sends a 'heartbeat' message after the timeout is passed to keep the http connection alive.
+    """).tag(config=True)
