@@ -1,5 +1,5 @@
 #############################################################################
-# Copyright (c) 2018, Voila Contributors                                    #
+# Copyright (c) 2018, Voilà Contributors                                    #
 # Copyright (c) 2018, QuantStack                                            #
 #                                                                           #
 # Distributed under the terms of the BSD 3-Clause License.                  #
@@ -13,17 +13,14 @@ import traitlets
 from traitlets.config import Config
 
 from jinja2 import contextfilter
+import jinja2
 
 from nbconvert.filters.markdown_mistune import IPythonRenderer, MarkdownWithMath
 from nbconvert.exporters.html import HTMLExporter
 from nbconvert.exporters.templateexporter import TemplateExporter
 from nbconvert.filters.highlight import Highlight2HTML
 
-from .threading import async_generator_to_thread
-
-# As long as we support Python35, we use this library to get as async
-# generators: https://pypi.org/project/async_generator/
-from async_generator import async_generator, yield_
+from .static_file_handler import TemplateStaticFileHandler
 
 
 class VoilaMarkdownRenderer(IPythonRenderer):
@@ -42,10 +39,14 @@ class VoilaMarkdownRenderer(IPythonRenderer):
 class VoilaExporter(HTMLExporter):
     """Custom HTMLExporter that inlines the images using VoilaMarkdownRenderer"""
 
+    base_url = traitlets.Unicode(help="Base url for resources").tag(config=True)
     markdown_renderer_class = traitlets.Type('mistune.Renderer').tag(config=True)
+    # Can be a ContentsManager from notebook or jupyter_server, so Any will have to do for now
+    contents_manager = traitlets.Any()
 
     # The voila exporter overrides the markdown renderer from the HTMLExporter
     # to inline images.
+
     @contextfilter
     def markdown2html(self, context, source):
         cell = context['cell']
@@ -61,9 +62,6 @@ class VoilaExporter(HTMLExporter):
     @property
     def default_config(self):
         c = Config({
-            'CSSHTMLHeaderPreprocessor': {
-                'enabled': False
-            },
             'VoilaExporter': {
                 'markdown_renderer_class': 'voila.exporter.VoilaMarkdownRenderer'
             }
@@ -71,19 +69,12 @@ class VoilaExporter(HTMLExporter):
         c.merge(super(VoilaExporter, self).default_config)
         return c
 
-    # Instead, we use the VoilaCSSPreprocessor.
-
-    @traitlets.default('preprocessors')
-    def _default_preprocessors(self):
-        return ['voila.csspreprocessor.VoilaCSSPreprocessor']
-
     # Overriding the default template file.
 
     @traitlets.default('template_file')
     def default_template_file(self):
-        return 'voila.tpl'
+        return 'index.html.j2'
 
-    @async_generator
     async def generate_from_notebook_node(self, nb, resources=None, extra_context={}, **kw):
         # this replaces from_notebook_node, but calls template.generate instead of template.render
         langinfo = nb.metadata.get('language_info', {})
@@ -106,25 +97,50 @@ class VoilaExporter(HTMLExporter):
                 'no_prompt': self.exclude_input_prompt and self.exclude_output_prompt,
                 }
 
-        # Jinja with Python3.5 does not support async (generators), which
-        # means that it's not async all the way down. Which means that we
-        # cannot use coroutines for the cell_generator, and that they will
-        # block the IO loop. In that case we will run the iterator in a
-        # thread instead.
-
-        @async_generator_to_thread
-        @async_generator
-        async def async_jinja_generator():
-            # Top level variables are passed to the template_exporter here.
-            for output in self.template.generate(nb=nb_copy, resources=resources, **extra_context):
-                await yield_((output, resources))
-
-        async for output, resources in async_jinja_generator():
-            await yield_((output, resources))
+        async for output in self.template.generate_async(nb=nb_copy, resources=resources, **extra_context, static_url=self.static_url):
+            yield (output, resources)
 
     @property
     def environment(self):
+        # enable Jinja async template execution
+        self.enable_async = True
         env = super(type(self), self).environment
         if 'jinja2.ext.do' not in env.extensions:
             env.add_extension('jinja2.ext.do')
         return env
+
+    def get_template_paths(self):
+        return self.template_path
+
+    def static_url(self, path):
+        """Mimics tornado.web.RequestHandler.static_url"""
+        settings = {
+            'static_url_prefix': f'{self.base_url}voila/templates/',
+            'static_path': None  # not used in TemplateStaticFileHandler.get_absolute_path
+        }
+        return TemplateStaticFileHandler.make_static_url(settings, f'{self.template_name}/static/{path}')
+
+    def _init_resources(self, resources):
+        def make_url(path):
+            # similar to static_url, but does not assume the static prefix
+            settings = {
+                'static_url_prefix': f'{self.base_url}voila/templates/',
+                'static_path': None  # not used in TemplateStaticFileHandler.get_absolute_path
+            }
+            return TemplateStaticFileHandler.make_static_url(settings, f'{self.template_name}/{path}')
+
+        def include_css(name):
+            code = f'<link rel="stylesheet" type="text/css" href="{make_url(name)}">'
+            return jinja2.Markup(code)
+
+        def include_js(name):
+            code = f'<script src="{make_url(name)}"></script>'
+            return jinja2.Markup(code)
+
+        def include_url(name):
+            return jinja2.Markup(make_url(name))
+        resources = super(VoilaExporter, self)._init_resources(resources)
+        resources['include_css'] = include_css
+        resources['include_js'] = include_js
+        resources['include_url'] = include_url
+        return resources
