@@ -1,5 +1,5 @@
 #############################################################################
-# Copyright (c) 2018, Voila Contributors                                    #
+# Copyright (c) 2018, Voilà Contributors                                    #
 # Copyright (c) 2018, QuantStack                                            #
 #                                                                           #
 # Distributed under the terms of the BSD 3-Clause License.                  #
@@ -7,7 +7,10 @@
 # The full license is in the file LICENSE, distributed with this software.  #
 #############################################################################
 
+import asyncio
 import os
+import sys
+import traceback
 
 import tornado.web
 
@@ -17,16 +20,21 @@ from jupyter_server.utils import url_path_join
 import nbformat
 
 from nbconvert.preprocessors import ClearOutputPreprocessor
+from nbclient.exceptions import CellExecutionError
+from nbclient.util import ensure_async
+from tornado.httputil import split_host_and_port
 
-from .execute import executenb, VoilaExecutePreprocessor
+from ._version import __version__
+from .execute import VoilaExecutor, strip_code_cell_warnings
 from .exporter import VoilaExporter
+from .paths import collect_template_paths
 
 
 class VoilaHandler(JupyterHandler):
 
     def initialize(self, **kwargs):
         self.notebook_path = kwargs.pop('notebook_path', [])    # should it be []
-        self.nbconvert_template_paths = kwargs.pop('nbconvert_template_paths', [])
+        self.template_paths = kwargs.pop('template_paths', [])
         self.traitlet_config = kwargs.pop('config', None)
         self.voila_configuration = kwargs['voila_configuration']
         # we want to avoid starting multiple kernels due to template mistakes
@@ -44,7 +52,7 @@ class VoilaHandler(JupyterHandler):
             # generate a list of nbextensions that are enabled for the classical notebook
             # a template can use that to load classical notebook extensions, but does not have to
             notebook_config = self.config_manager.get('notebook')
-            # except for the widget extension itself, since voila has its own
+            # except for the widget extension itself, since Voilà has its own
             load_extensions = notebook_config.get('load_extensions', {})
             if 'jupyter-js-widgets/extension' in load_extensions:
                 load_extensions['jupyter-js-widgets/extension'] = False
@@ -54,19 +62,47 @@ class VoilaHandler(JupyterHandler):
         else:
             nbextensions = []
 
-        self.notebook = await self.load_notebook(notebook_path)
-        if not self.notebook:
+        notebook = await self.load_notebook(notebook_path)
+        if not notebook:
             return
         self.cwd = os.path.dirname(notebook_path)
 
         path, basename = os.path.split(notebook_path)
         notebook_name = os.path.splitext(basename)[0]
 
+        # Adding request uri to kernel env
+        self.kernel_env = os.environ.copy()
+        self.kernel_env['SCRIPT_NAME'] = self.request.path
+        self.kernel_env['PATH_INFO'] = ''  # would be /foo/bar if voila.ipynb/foo/bar was supported
+        self.kernel_env['QUERY_STRING'] = str(self.request.query)
+        self.kernel_env['SERVER_SOFTWARE'] = 'voila/{}'.format(__version__)
+        self.kernel_env['SERVER_PROTOCOL'] = str(self.request.version)
+        host, port = split_host_and_port(self.request.host.lower())
+        self.kernel_env['SERVER_PORT'] = str(port) if port else ''
+        self.kernel_env['SERVER_NAME'] = host
+
+        # we can override the template via notebook metadata or a query parameter
+        template_override = None
+        if 'voila' in notebook.metadata and self.voila_configuration.allow_template_override in ['YES', 'NOTEBOOK']:
+            template_override = notebook.metadata['voila'].get('template')
+        if self.voila_configuration.allow_template_override == 'YES':
+            template_override = self.get_argument("voila-template", template_override)
+        if template_override:
+            self.template_paths = collect_template_paths(['voila', 'nbconvert'], template_override)
+        template_name = template_override or self.voila_configuration.template
+
+        theme = self.voila_configuration.theme
+        if 'voila' in notebook.metadata and self.voila_configuration.allow_theme_override in ['YES', 'NOTEBOOK']:
+            theme = notebook.metadata['voila'].get('theme', theme)
+        if self.voila_configuration.allow_theme_override == 'YES':
+            theme = self.get_argument("voila-theme", theme)
+
         # render notebook to html
         resources = {
             'base_url': self.base_url,
             'nbextensions': nbextensions,
-            'theme': self.voila_configuration.theme,
+            'theme': theme,
+            'template': template_name,
             'metadata': {
                 'name': notebook_name
             }
@@ -76,15 +112,23 @@ class VoilaHandler(JupyterHandler):
         extra_resources = self.voila_configuration.config.VoilaConfiguration.resources
         # if no resources get configured from neither the CLI nor a config file,
         # extra_resources is a traitlets.config.loader.LazyConfigValue object
+        # This seems to only happy with the notebook server and traitlets 5
+        # Note that we use string checking for backward compatibility
+        if 'DeferredConfigString' in str(type(extra_resources)):
+            from .configuration import VoilaConfiguration
+            extra_resources = VoilaConfiguration.resources.from_string(extra_resources)
         if not isinstance(extra_resources, dict):
             extra_resources = extra_resources.to_dict()
         if extra_resources:
             recursive_update(resources, extra_resources)
 
         self.exporter = VoilaExporter(
-            template_path=self.nbconvert_template_paths,
+            template_paths=self.template_paths,
+            template_name=template_name,
             config=self.traitlet_config,
-            contents_manager=self.contents_manager  # for the image inlining
+            contents_manager=self.contents_manager,  # for the image inlining
+            theme=theme,  # we now have the theme in two places
+            base_url=self.base_url,
         )
         if self.voila_configuration.strip_sources:
             self.exporter.exclude_input = True
@@ -96,20 +140,15 @@ class VoilaHandler(JupyterHandler):
         # Template should first call kernel_start, and then decide to use notebook_execute
         # or cell_generator to implement progressive cell rendering
         extra_context = {
-            # NOTE: we can remove the lambda is we use jinja's async feature, which will automatically await the future
-            'kernel_start': lambda: self._jinja_kernel_start().result(),  # pass the result (not the future) to the template
+            'kernel_start': self._jinja_kernel_start,
             'cell_generator': self._jinja_cell_generator,
             'notebook_execute': self._jinja_notebook_execute,
         }
 
-        # Currenly _jinja_kernel_start is executed from a different thread, which causes the websocket connection from
-        # the frontend to fail. Instead, we start it beforehand, and just return the kernel_id in _jinja_kernel_start
-        self.kernel_id = await tornado.gen.maybe_future(self.kernel_manager.start_kernel(kernel_name=self.notebook.metadata.kernelspec.name, path=self.cwd))
-
         # Compose reply
         self.set_header('Content-Type', 'text/html')
         # render notebook in snippets, and flush them out to the browser can render progresssively
-        async for html_snippet, resources in self.exporter.generate_from_notebook_node(self.notebook, resources=resources, extra_context=extra_context):
+        async for html_snippet, resources in self.exporter.generate_from_notebook_node(notebook, resources=resources, extra_context=extra_context):
             self.write(html_snippet)
             self.flush()  # we may not want to consider not flusing after each snippet, but add an explicit flush function to the jinja context
             # yield  # give control back to tornado's IO loop, so it can handle static files or other requests
@@ -118,35 +157,85 @@ class VoilaHandler(JupyterHandler):
     def redirect_to_file(self, path):
         self.redirect(url_path_join(self.base_url, 'voila', 'files', path))
 
-    @tornado.gen.coroutine
-    def _jinja_kernel_start(self):
+    async def _jinja_kernel_start(self, nb):
         assert not self.kernel_started, "kernel was already started"
-        # See command above aboout not being able to start the kernel from a different thread
-        self.kernel_started = True
-        return self.kernel_id
 
-    def _jinja_notebook_execute(self, nb, kernel_id):
+        kernel_id = await ensure_async(self.kernel_manager.start_kernel(
+           kernel_name=nb.metadata.kernelspec.name,
+           path=self.cwd,
+           env=self.kernel_env,
+        ))
         km = self.kernel_manager.get_kernel(kernel_id)
-        result = executenb(nb, km=km, cwd=self.cwd, config=self.traitlet_config)
+
+        self.executor = VoilaExecutor(nb, km=km, config=self.traitlet_config)
+
+        ###
+        # start kernel client
+        self.executor.kc = km.client()
+        await ensure_async(self.executor.kc.start_channels())
+        await ensure_async(self.executor.kc.wait_for_ready(timeout=self.executor.startup_timeout))
+        self.executor.kc.allow_stdin = False
+        ###
+
+        self.kernel_started = True
+        return kernel_id
+
+    async def _jinja_notebook_execute(self, nb, kernel_id):
+        result = await self.executor.async_execute(cleanup_kc=False)
         # we modify the notebook in place, since the nb variable cannot be reassigned it seems in jinja2
         # e.g. if we do {% with nb = notebook_execute(nb, kernel_id) %}, the base template/blocks will not
         # see the updated variable (it seems to be local to our block)
         nb.cells = result.cells
 
-    def _jinja_cell_generator(self, nb, kernel_id):
+    async def _jinja_cell_generator(self, nb, kernel_id):
         """Generator that will execute a single notebook cell at a time"""
-        km = self.kernel_manager.get_kernel(kernel_id)
-
         nb, resources = ClearOutputPreprocessor().preprocess(nb, {'metadata': {'path': self.cwd}})
-        ep = VoilaExecutePreprocessor(config=self.traitlet_config)
+        for cell_idx, input_cell in enumerate(nb.cells):
+            try:
+                task = asyncio.ensure_future(self.executor.execute_cell(input_cell, None, cell_idx, store_history=False))
+                while True:
+                    done, pending = await asyncio.wait({task}, timeout=self.voila_configuration.http_keep_alive_timeout)
+                    if pending:
+                        # If not done within the timeout, we send a heartbeat
+                        # this is fundamentally to avoid browser/proxy read-timeouts, but
+                        # can be used in a template to give feedback to a user
+                        self.write("<script>voila_heartbeat()</script>\n")
+                        self.flush()
+                        continue
+                    output_cell = await task
+                    break
+            except TimeoutError:
+                output_cell = input_cell
+                break
+            except CellExecutionError:
+                if self.executor.should_strip_error():
+                    strip_code_cell_warnings(input_cell)
+                    self.executor.strip_code_cell_errors(input_cell)
+                output_cell = input_cell
+                break
+            except Exception as e:
+                self.log.exception('Error at server while executing cell: %r', input_cell)
+                output_cell = nbformat.v4.new_code_cell()
+                if self.executor.should_strip_error():
+                    output_cell.outputs = [
+                        {
+                            "output_type": "stream",
+                            "name": "stderr",
+                            "text": "An exception occurred at the server (not the notebook). {}".format(self.executor.cell_error_instruction),
+                        }
+                    ]
+                else:
+                    output_cell.outputs = [
+                        {
+                            'output_type': 'error',
+                            'ename': type(e).__name__,
+                            'evalue': str(e),
+                            'traceback': traceback.format_exception(*sys.exc_info()),
+                        }
+                    ]
+            finally:
+                yield output_cell
 
-        with ep.setup_preprocessor(nb, resources, km=km):
-            for cell_idx, cell in enumerate(nb.cells):
-                res = ep.preprocess_cell(cell, resources, cell_idx, store_history=False)
-
-                yield res[0]
-
-    # @tornado.gen.coroutine
     async def load_notebook(self, path):
         model = self.contents_manager.get(path=path)
         if 'content' not in model:
